import inspect
from collections import OrderedDict
from supervisely.sly_logger import logger
from supervisely.annotation.annotation import Annotation
from supervisely.project.project_meta import ProjectMeta
try:
    from imgaug.augmentables.bbs import BoundingBox, BoundingBoxesOnImage
    from imgaug.augmentables.segmaps import SegmentationMapsOnImage
except:
    pass
import numpy as np


if not hasattr(np, 'bool'): np.bool = np.bool_

def create_aug_info(category_name, aug_name, params, sometimes: float = None):
    clean_params = params
    #clean_params = remove_unexpected_arguments(category_name, aug_name, params)
    res = {
        "category": category_name,
        "name": aug_name,
        "params": clean_params,
    }
    if sometimes is not None:
        if type(sometimes) is not float or not (0.0 <= sometimes <= 1.0):
            raise ValueError(f"sometimes={sometimes}, type != {type(sometimes)}")
        res["sometimes"] = sometimes
    res["python"] = aug_to_python(res)
    return res


def aug_to_python(aug_info):
    pstr = ""
    for name, value in aug_info["params"].items():
        v = value
        if isinstance(v, list) and len(v) == 2:  #name != 'nb_iterations' and
            v = tuple(v)
        elif isinstance(v, dict) and "x" in v.keys() and "y" in v.keys():
            x = v["x"]
            y = v["y"]
            if isinstance(x, list) and len(x) == 2:
                x = tuple(x)
            if isinstance(y, list) and len(y) == 2:
                y = tuple(y)
            v = {"x": x, "y": y}
            
        if type(value) is str:
            pstr += f"{name}='{v}', "
        else:
            pstr += f"{name}={v}, "
    method_py = f"iaa.{aug_info['category']}.{aug_info['name']}({pstr[:-2]})"

    res = method_py
    if "sometimes" in aug_info:
        res = f"iaa.Sometimes({aug_info['sometimes']}, {method_py})"
    return res


def pipeline_to_python(aug_infos, random_order=False):
    template = \
"""import imgaug.augmenters as iaa

seq = iaa.Sequential([
{}
], random_order={})
"""
    py_lines = []
    for info in aug_infos:
        line = aug_to_python(info)
        _validate = info["python"]
        if line != _validate:
            raise ValueError("Generated python line differs from the one from config: \n\n{!r}\n\n{!r}"
                             .format(line, _validate))
        py_lines.append(line)
    res = template.format('\t' + ',\n\t'.join(py_lines), random_order)
    return res


def get_default_params_by_function(f):
    params = []
    method_info = inspect.signature(f)
    for param in method_info.parameters.values():
        formatted = str(param)
        if 'deprecated' in formatted or 'seed=None' in formatted or 'name=None' in formatted:
            continue
        if param.default == inspect.Signature.empty:
            continue
        params.append({
            "pname": param.name,
            "default": param.default
        })
    return params


def get_default_params_by_name(category_name, aug_name):
    func = get_function(category_name, aug_name)
    defaults = get_default_params_by_function(func)
    return defaults


def get_function(category_name, aug_name):
    try:
        import imgaug.augmenters as iaa
    except ModuleNotFoundError as e:
        logger.error(f'{e}. Try to install extra dependencies. Run "pip install supervisely[aug]"')
        raise e
    try:
        submodule = getattr(iaa, category_name)
        aug_f = getattr(submodule, aug_name)
        return aug_f
    except Exception as e:
        logger.error(repr(e))
        # raise e
        return None


def build_pipeline(aug_infos, random_order=False):
    try:
        import imgaug.augmenters as iaa
    except ModuleNotFoundError as e:
        logger.error(f'{e}. Try to install extra dependencies. Run "pip install supervisely[aug]"')
        raise e
    pipeline = []
    for aug_info in aug_infos:
        category_name = aug_info["category"]
        aug_name = aug_info["name"]
        params = aug_info["params"]

        aug_func = get_function(category_name, aug_name)

        for param_name, param_val in params.items():
            if isinstance(param_val, dict):
                if "x" in param_val.keys() and "y" in param_val.keys():
                    if isinstance(param_val["x"], list) and len(param_val["x"]) == 2:
                        param_val["x"] = tuple(param_val["x"])
                    if isinstance(param_val["y"], list) and len(param_val["y"]) == 2:
                        param_val["y"] = tuple(param_val["y"])
            elif isinstance(param_val, list) and len(param_val) == 2:
                # all {'par': [n1, n2]} to {'par': (n1, n2)}
                params[param_name] = tuple(param_val)

        aug = aug_func(**params)

        sometimes = aug_info.get("sometimes", None)
        if sometimes is not None:
            aug = iaa.meta.Sometimes(sometimes, aug)
        pipeline.append(aug)
    augs = iaa.Sequential(pipeline, random_order=random_order)
    return augs


def build(aug_info):
    return build_pipeline([aug_info])


def remove_unexpected_arguments(category_name, aug_name, params):
    # to avoid this:
    # TypeError: f() got an unexpected keyword argument 'b'
    defaults = get_default_params_by_name(category_name, aug_name)
    allowed_names = [d["pname"] for d in defaults]

    res = OrderedDict()
    for name, value in params.items():
        if name in allowed_names:
            res[name] = value
    return res


def _apply(augs, img, boxes=None, masks=None):
    try:
        import imgaug.augmenters as iaa
    except ModuleNotFoundError as e:
        logger.error(f'{e}. Try to install extra dependencies. Run "pip install supervisely[aug]"')
        raise e
    augs: iaa.Sequential
    res = augs(images=[img], bounding_boxes=boxes, segmentation_maps=masks)
    #return image, boxes, masks
    return res[0][0], res[1], res[2]


def apply(augs, meta: ProjectMeta, img, ann: Annotation, segmentation_type='semantic'):
    # @TODO: save object tags

    # works for rectangles
    det_meta, det_mapping = meta.to_detection_task(convert_classes=False)
    det_ann = ann.to_detection_task(det_mapping)
    ia_boxes = det_ann.bboxes_to_imgaug()

    # works for polygons and bitmaps
    seg_meta, seg_mapping = meta.to_segmentation_task()
    seg_ann = ann.to_nonoverlapping_masks(seg_mapping)

    index_to_class = {}
    if segmentation_type == 'semantic':
        seg_ann = seg_ann.to_segmentation_task()
        class_to_index = {obj_class.name: idx for idx, obj_class in enumerate(seg_meta.obj_classes, start=1)}
        index_to_class = {v: k for k, v in class_to_index.items()}
    elif segmentation_type == 'instance':
        class_to_index = None
        index_to_class = {idx: label.obj_class.name for idx, label in enumerate(seg_ann.labels, start=1)}
    elif segmentation_type == 'panoptic':
        raise NotImplementedError

    ia_masks = seg_ann.masks_to_imgaug(class_to_index)

    res_meta = det_meta.merge(seg_meta)  # TagMetas should be preserved

    res_img, res_ia_boxes, res_ia_masks = _apply(augs, img, ia_boxes, ia_masks)
    res_ann = Annotation.from_imgaug(res_img,
                                     ia_boxes=res_ia_boxes, ia_masks=res_ia_masks,
                                     index_to_class=index_to_class, meta=res_meta)
    # add image tags
    res_ann = res_ann.clone(img_tags=ann.img_tags)
    return res_meta, res_img, res_ann


def apply_to_image(augs, img):
    res_img, _, _ = _apply(augs, img, None, None)
    return res_img


def _instances_to_nonoverlapping_mask(instance_masks):
    """Convert instance segmentation masks to nonoverlapping objects on
    semantic-like segmentation mask.


    :param instance_masks: Instance segmentation masks.
    :type instance_masks: (H,W,C) ndarray, UINT
    :returns: Segmentation map with all instances.
    :rtype: (H,W) ndarray
    """
    common_img = np.zeros(instance_masks.shape[:2], np.int32)
    n_instances = instance_masks.shape[2]

    for idx in range(n_instances):
        common_img[instance_masks[:,:,idx].astype(np.bool)] = idx + 1

    return common_img

def _nonoverlapping_mask_to_instances(common_mask, n_instances):
    """Convert nonoverlapping objects on semantic-like segmentation mask to
    instance segmentation masks.

    :param common_mask: Segmentation map with all instances.
    :type common_mask: (H,W) ndarray, UINT
    :param n_instances: Number of objects on segmentation map. After applying augmentations some of instances
        may disappear from image. In this case result array may contain empty channels.
    :type n_instances: int
    :returns: Instance segmentation masks.
    :rtype: (H,W,C) ndarray
    """
    bitmap_masks = []
    for idx in range(n_instances):
        instance_mask = common_mask == idx + 1
        bitmap_masks.append(instance_mask.astype(np.uint8))
    return np.stack(bitmap_masks, axis=-1)


def apply_to_image_and_bbox(augs, img, bboxes):
    """Apply augmentations to image and bounding boxes.

    :param augs: Defined imgaug augmentation sequence.
    :type augs: iaa.Sequential
    :param img: Input image.
    :type img: (H, W) or (H, W, C) ndarray
    :param bboxes: Bounding boxes.
    :type bboxes: List[bbox], where bbox: list or tuple of bbox coords in XYXY format
    :returns: Augmented image.
    :rtype: (H,W,C) ndarray
    :returns: Augmented bounding boxes.
    :rtype: List[bbox], where bbox: list of bbox coords in XYXY format
    """
    assert isinstance(bboxes, list)

    ia_boxes = [BoundingBox(box[0], box[1], box[2], box[3]) for box in bboxes]
    ia_boxes = BoundingBoxesOnImage(ia_boxes, shape=img.shape[:2])
    res_img, res_boxes, _ = _apply(augs, img, boxes=ia_boxes)
    res_boxes = [[res_box.x1, res_box.y1, res_box.x2, res_box.y2] for res_box in res_boxes]

    return res_img, res_boxes


def apply_to_image_and_mask(augs, img, mask, segmentation_type="semantic"):
    """Apply augmentations to image and segmentation mask (instance or semantic).

    :param augs: Defined imgaug augmentation sequence.
    :type augs: iaa.Sequential
    :param img: Input image.
    :type img: (H, W) or (H, W, C) ndarray
    :param mask: Instance or semantic segmentation mask.
    :type mask: (H, W) ndarray if semantic or (H, W, C) ndarray if instance
    :param segmentation_type: Define how to process segmentation masks.
    :type segmentation_type: str, one of ('semantic', 'instance')
    :returns: Augmented image.
    :rtype: (H,W,C) ndarray
    :returns: Augmented segmentation mask.
    :rtype: (H,W) ndarray if semantic or (H,W,C) ndarray if instance
    (H,W) ndarray if semantic or (H,W,C) ndarray if instance
        Augmented segmentation mask.
    :rtype: List[bbox], where bbox: list of bbox coords in XYXY format
    """
    assert isinstance(segmentation_type, str) and segmentation_type in ["semantic", "instance"]
    assert isinstance(mask, np.ndarray)
    assert mask.shape[:2] == img.shape[:2]

    if segmentation_type == "instance":
        N_instances = mask.shape[2]
        mask = _instances_to_nonoverlapping_mask(mask) # [H,W,C] -> [H,W]

    segmaps = SegmentationMapsOnImage(mask, shape=img.shape[:2])

    res_img, _, res_segmaps = _apply(augs, img, masks=segmaps)

    res_mask = res_segmaps.get_arr()
    if segmentation_type == "instance":
        res_mask = _nonoverlapping_mask_to_instances(res_mask, N_instances) # [H,W] -> [H,W,C]

    if res_img.shape[:2] != res_mask.shape[:2]:
        raise ValueError(f"Image and mask have different shapes "
                         f"({res_img.shape[:2]} != {res_mask.shape[:2]}) after augmentations. "
                         f"Please, contact tech support")
    return res_img, res_mask


def apply_to_image_bbox_and_mask(augs, img, bboxes, mask, segmentation_type="semantic"):
    """Apply augmentations to image, bounding boxes and segmentation mask
    (instance or semantic).

    :param augs: Defined imgaug augmentation sequence.
    :type augs: iaa.Sequential
    :param img: Input image.
    :type img: (H, W) or (H, W, C) ndarray
    :param bboxes: Bounding boxes.
    :type bboxes: List[bbox], where bbox: list or tuple of bbox coords in XYXY format
    :param mask: Instance or semantic segmentation mask.
    :type mask: (H, W) ndarray if semantic or (H, W, C) ndarray if instance
    :param segmentation_type: Define how to process segmentation masks.
    :type segmentation_type: str, one of ('semantic', 'instance')
    :returns: Augmented image.
    :rtype: (H,W,C) ndarray
    :returns: Augmented bounding boxes.
    :rtype: List[bbox], where bbox: list of bbox coords in XYXY format
    :returns: Augmented segmentation mask.
    :rtype: (H,W) ndarray if semantic or (H,W,C) ndarray if instance
    """
    assert isinstance(bboxes, list)
    assert isinstance(segmentation_type, str) and segmentation_type in ["semantic", "instance"]
    assert isinstance(mask, np.ndarray)
    assert img.shape[:2] == mask.shape[:2]

    if segmentation_type == "semantic":
        assert mask.ndim == 2
    elif segmentation_type == "instance":
        assert mask.ndim == 3

    boxes = [BoundingBox(box[0], box[1], box[2], box[3]) for box in bboxes]
    boxes = BoundingBoxesOnImage(boxes, shape=img.shape[:2])

    if segmentation_type == "instance":
        N_instances = mask.shape[2]
        mask = _instances_to_nonoverlapping_mask(mask) # [H,W,C] -> [H,W]
    segmaps = SegmentationMapsOnImage(mask, shape=mask.shape)

    res_img, res_boxes, res_segmaps = _apply(augs, img, boxes=boxes, masks=segmaps)

    res_mask = res_segmaps.get_arr()
    if segmentation_type == "instance":
        res_mask = _nonoverlapping_mask_to_instances(res_mask, N_instances) # [H,W] -> [H,W,C]
    res_boxes = [[res_box.x1, res_box.y1, res_box.x2, res_box.y2] for res_box in res_boxes]

    return res_img, res_boxes, res_mask


def apply_to_image_bbox_and_both_types_masks(augs, img, bboxes, semantic_mask, instance_masks):
    """
    Apply augmentations to image, bounding boxes and both types of segmentation masks (instance and semantic) together.

    :param augs: Defined imgaug augmentation sequence.
    :type augs: iaa.Sequential
    :param img: Input image.
    :type img: (H, W) or (H, W, C) ndarray
    :param bboxes: Bounding boxes.
    :type bboxes: List[bbox], where bbox: list or tuple of bbox coords in XYXY format
    :param semantic_mask: Semantic segmentation mask.
    :type semantic_mask: (H,W) ndarray
    :param instance_masks: Instance segmentation masks.
    :type instance_masks: (H,W,C) ndarray
    :returns: Augmented image, augmented bounding boxes, augmented semantic segmentation mask, augmented instance segmentation masks
    :rtype: (H,W) or (H,W,C) ndarray, List[bbox], (H,W) ndarray, (H,W,C) ndarray
    """
    assert isinstance(bboxes, list)
    assert isinstance(semantic_mask, np.ndarray) and semantic_mask.ndim == 2
    assert isinstance(instance_masks, np.ndarray) and instance_masks.ndim == 3
    assert img.shape[:2] == semantic_mask.shape[:2] == instance_masks.shape[:2]

    boxes = [BoundingBox(box[0], box[1], box[2], box[3]) for box in bboxes]
    boxes = BoundingBoxesOnImage(boxes, shape=img.shape[:2])

    N_instances = instance_masks.shape[2]
    merged_instance_masks = _instances_to_nonoverlapping_mask(instance_masks) # [H,W,C] -> [H,W]
    # merge instance and segmentation masks as different channels of one mask
    all_masks = np.stack((merged_instance_masks, semantic_mask, semantic_mask), axis=-1)
    segmaps = SegmentationMapsOnImage(all_masks, shape=all_masks.shape)

    res_img, res_boxes, res_segmaps = _apply(augs, img, boxes=boxes, masks=segmaps)

    res_masks = res_segmaps.get_arr()
    res_semantic = res_masks[:,:,2]
    res_instances = _nonoverlapping_mask_to_instances(res_masks[:,:,0], N_instances) # [H,W] -> [H,W,C]
    res_boxes = [[res_box.x1, res_box.y1, res_box.x2, res_box.y2] for res_box in res_boxes]

    return res_img, res_boxes, res_semantic, res_instances



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from typing import Callable, Optional

from supervisely.app import DataJson
from supervisely.app.widgets import Widget


class Button(Widget):
    """Clickable button with configurable text, style, and icon."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        CLICK = "button_clicked_cb"

    def __init__(
        self,
        text: Optional[str] = "Button",
        button_type: Optional[
            Literal["primary", "info", "warning", "danger", "success", "text"]
        ] = "primary",
        button_size: Optional[Literal["mini", "small", "large"]] = None,
        plain: Optional[bool] = False,
        show_loading: Optional[bool] = True,
        icon: Optional[str] = None,
        icon_gap: Optional[int] = 5,
        widget_id: Optional[str] = None,
        link: Optional[str] = None,
        emit_on_click: Optional[str] = None,
        style: Optional[str] = None,
        call_on_click: Optional[str] = None,
        visible_by_vue_field: Optional[str] = "",
        icon_color: Optional[str] = None,
    ):
        """
        :param text: Button label.
        :type text: str, optional
        :param button_type: primary, info, warning, danger, success, or text.
        :type button_type: Literal, optional
        :param button_size: mini, small, or large.
        :type button_size: Literal, optional
        :param plain: Plain style (no fill).
        :type plain: bool, optional
        :param show_loading: Show loading spinner when processing.
        :type show_loading: bool, optional
        :param icon: Icon class (e.g. "zmdi zmdi-play").
        :type icon: str, optional
        :param icon_gap: Gap between icon and text (px).
        :type icon_gap: int, optional
        :param widget_id: Widget identifier.
        :type widget_id: str, optional
        :param link: URL to open on click.
        :type link: str, optional
        :param emit_on_click: Event name to emit on click.
        :type emit_on_click: str, optional
        :param style: Inline CSS.
        :type style: str, optional
        :param call_on_click: Callback name for click.
        :type call_on_click: str, optional
        :param icon_color: Icon color.
        :type icon_color: str, optional

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import Button
                button = Button(text="Run", button_type="primary", button_size="large")
        """
        self._widget_routes = {}

        self._text = text
        self._button_type = button_type
        self._button_size = button_size
        self._plain = plain
        self._icon_gap = icon_gap
        self._link = link
        if icon is None:
            self._icon = ""
        else:
            icon_style = f"margin-right: {icon_gap}px"
            if icon_color is not None:
                icon_style += f"; color: {icon_color}"
            self._icon = f'<i class="{icon}" style="{icon_style}"></i>'

        self._loading = False
        self._disabled = False
        self._show_loading = show_loading
        self._click_handled = False
        self._emit_on_click = emit_on_click
        self._style = style
        self._call_on_click = call_on_click
        self._visible_by_vue_field = visible_by_vue_field

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - text: Text to be displayed on the button.
            - button_type: Type of the button.
            - plain: If True, the button will be plain.
            - button_size: Size of the button.
            - loading: If True, the button will show loading animation.
            - disabled: If True, the button will be disabled.
            - icon: Icon to be displayed on the button.
            - link: Link to be opened on button click.
        """
        return {
            "text": self._text,
            "button_type": self._button_type,
            "plain": self._plain,
            "button_size": self._button_size,
            "loading": self._loading,
            "disabled": self._disabled,
            "icon": self._icon,
            "link": self._link,
            "style": self._style,
        }

    def get_json_state(self) -> None:
        """Button widget doesn't have state, so this method returns None."""
        return {"visible_by_vue_field": self._visible_by_vue_field}

    @property
    def text(self) -> str:
        """Returns the text to be displayed on the button.

        :returns: Text to be displayed on the button.
        :rtype: str
        """
        return self._text

    @text.setter
    def text(self, value: str) -> None:
        """Sets the text to be displayed on the button.

        :param value: Text to be displayed on the button.
        :type value: str
        """
        self._text = value
        DataJson()[self.widget_id]["text"] = self._text
        DataJson().send_changes()

    @property
    def icon(self) -> str:
        """Returns the icon to be displayed on the button.

        :returns: Icon to be displayed on the button.
        :rtype: str
        """
        return self._icon

    @icon.setter
    def icon(self, value: str) -> None:
        """Sets the icon to be displayed on the button.

        :param value: Icon to be displayed on the button.
        :type value: str
        """
        if value is None:
            self._icon = ""
        else:
            self._icon = f'<i class="{value}" style="margin-right: {self._icon_gap}px"></i>'
        DataJson()[self.widget_id]["icon"] = self._icon
        DataJson().send_changes()

    @property
    def button_type(self) -> str:
        """Returns the type of the button.

        :returns: Type of the button.
        :rtype: str
        """
        return self._button_type

    @button_type.setter
    def button_type(
        self, value: Literal["primary", "info", "warning", "danger", "success", "text"]
    ) -> None:
        """Sets the type of the button.

        :param value: Type of the button.
        :type value: Literal["primary", "info", "warning", "danger", "success", "text"]
        """
        self._button_type = value
        DataJson()[self.widget_id]["button_type"] = self._button_type
        DataJson().send_changes()

    @property
    def plain(self) -> bool:
        """Returns True if the button is plain, False otherwise.

        :returns: True if the button is plain, False otherwise.
        :rtype: bool
        """
        return self._plain

    @plain.setter
    def plain(self, value: bool) -> None:
        """Sets the button to be plain or not.

        :param value: If True, the button will be plain.
        :type value: bool
        """
        self._plain = value
        DataJson()[self.widget_id]["plain"] = self._plain
        DataJson().send_changes()

    @property
    def link(self) -> str:
        """Returns the link to be opened on button click.

        :returns: Link to be opened on button click.
        :rtype: str
        """
        return self._link

    @link.setter
    def link(self, value: str) -> None:
        """Sets the link to be opened on button click.

        :param value: Link to be opened on button click.
        :type value: str
        """
        self._link = value
        DataJson()[self.widget_id]["link"] = self._link
        DataJson().send_changes()

    @property
    def loading(self) -> bool:
        """Returns True if the button shows loading animation, False otherwise.

        :returns: True if the button shows loading animation, False otherwise.
        :rtype: bool
        """
        return self._loading

    @loading.setter
    def loading(self, value: bool) -> None:
        """Sets the button loading animation.

        :param value: If True, the animation will be enabled, otherwise disabled.
        :type value: bool
        """
        self._loading = value
        DataJson()[self.widget_id]["loading"] = self._loading
        DataJson().send_changes()

    @property
    def show_loading(self) -> bool:
        """Returns True if the button shows loading animation, False otherwise.

        :returns: True if the button shows loading animation, False otherwise.
        :rtype: bool
        """
        return self._show_loading

    @property
    def disabled(self) -> bool:
        """Returns True if the button is disabled, False otherwise.

        :returns: True if the button is disabled, False otherwise.
        :rtype: bool
        """
        return self._disabled

    @property
    def style(self) -> Optional[str]:
        """Returns the CSS style applied to the button.

        :returns: CSS style applied to the button.
        :rtype: Optional[str]
        """
        return self._style

    @style.setter
    def style(self, value: Optional[str]) -> None:
        """Sets the CSS style to be applied to the button.

        :param value: CSS style to be applied to the button.
        :type value: Optional[str]
        """
        self._style = value
        DataJson()[self.widget_id]["style"] = self._style
        DataJson().send_changes()

    @disabled.setter
    def disabled(self, value: bool) -> None:
        """Sets the button to be disabled or not.

        :param value: If True, the button will be disabled.
        :type value: bool
        """
        self._disabled = value
        DataJson()[self.widget_id]["disabled"] = self._disabled

    def click(self, func: Callable[[], None]) -> Callable[[], None]:
        """Decorator that allows to handle button click. Decorated function
        will be called on button click.

        :param func: Function to be called on button click.
        :type func: Callable
        :returns: Decorated function.
        :rtype: Callable
        """
        route_path = self.get_route_path(Button.Routes.CLICK)
        server = self._sly_app.get_server()
        self._click_handled = True

        @server.post(route_path)
        def _click():
            if self.show_loading:
                self.loading = True
            try:
                func()
            except Exception as e:
                if self.show_loading and self.loading:
                    self.loading = False
                raise e
            if self.show_loading:
                self.loading = False

        return _click

    def _get_on_click(self):
        on_click_actions = []
        if self._emit_on_click:
            on_click_actions.append(f"$emit('{self._emit_on_click}');")

        if self._call_on_click:
            on_click_actions.append(f"{self._call_on_click}")

        if self._click_handled:
            on_click_actions.append(f"post('/{self.widget_id}/button_clicked_cb');")
        return " ".join(on_click_actions) if on_click_actions else None
